/*

  Doom 64 Macro Compiler

  Copyright 2008 James Haley

  haleyjd@hotmail.com

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <fstream>
#include <exception>
#include <cstdlib>

#include "macrodata.h"
#include "linetable.h"

using namespace std;

//==============================================================================
//
// Global Data
//

MacroHeader  header; // header data
Macro       *macros; // the macros

//==============================================================================
//
// Macro Reading
//

//
// ReadSint16
//
// Read a 16-bit little-endian integer from the buffer and advance the pointer.
//
static int ReadSint16(byte **buffer)
{
   int ret;

   ret  = (int)((*buffer)[0]);
   ret |= (int)((*buffer)[1]) << 8;

   *buffer += 2;

   // perform sign extension
   if(ret & 0x8000)
      ret = -ret;

   return ret;
}

//
// ReadMacroHeader
//
// The MACROS lump header is two 16-bit ints; the number of macros, and the
// total number of commands in all macros.
//
static void ReadMacroHeader(byte **buffer)
{
   header.macrocount   = ReadSint16(buffer);
   header.commandcount = ReadSint16(buffer);
}

//
// ReadMacro
//
// Reads an individual macro from the buffer.
//
// Macro binary format:
//   count [2 bytes]: number of commands in this macro
//   followed by count + 1 *
//      batchid [2 bytes]: batch # of this command
//      linetag [2 bytes]: line tag or special data
//      special [2 bytes]: line special
//
// There is a final junk command at the end of each macro.
// This command is not included in the count of commands.
//
static void ReadMacro(Macro *macro, byte **buffer)
{
   int i;
   MacroCommand *cmd;

   macro->commandcount = ReadSint16(buffer);

   // allocate commands
   macro->cmds = new MacroCommand [macro->commandcount];

   // read commands
   for(i = 0; i < macro->commandcount; ++i)
   {
      cmd = &(macro->cmds[i]);

      cmd->batchid     = ReadSint16(buffer);
      cmd->linetag     = ReadSint16(buffer);
      cmd->linespecial = ReadSint16(buffer);
   }

   // skip the junk command at the end
   *buffer += 3*2;
}

void LoadMacros(const char *fn)
{
   int i;
   ifstream infile;
   int length;
   byte  *buffer = NULL, *bufptr = NULL;

   infile.open(fn, ios::binary);

   if(!infile.is_open())
      throw exception("Failed to open macro file.");

   // get length of file
   infile.seekg(0, ios::end);
   length = infile.tellg();
   infile.seekg(0, ios::beg);

   // ensure meets minimum valid length
   if(length < 4)
      throw exception("Macro file is empty.");

   // allocate buffer
   bufptr = buffer = new byte [length];

   // read file into buffer
   infile.read((char *)buffer, length);

   // done with input file
   infile.close();

   // read header
   ReadMacroHeader(&bufptr);

   // check size of buffer for consistency
   if(length < 2 * (2 + header.macrocount 
                      + (header.macrocount + header.commandcount) * 3))
      throw exception("Corrupt macro file: too small!");

   // allocate macros
   macros = new Macro [header.macrocount];

   // read macros!
   for(i = 0; i < header.macrocount; ++i)
      ReadMacro(&macros[i], &bufptr);

   // done with input buffer
   delete [] buffer;
}

//
// DecompileMacros
//
// Call to write out loaded binary macros to a script file.
//
void DecompileMacros(const char *fn)
{
   int i;
   int curbatch  = -1;
   ofstream outfile;

   outfile.open(fn);

   if(!outfile.is_open())
      throw exception("Could not open script output file");

   for(i = 0; i < header.macrocount; ++i)
   {
      int j;
      Macro *macro = &macros[i];

      outfile << "macro" << endl;

      for(j = 0; j < macro->commandcount; ++j)
      {
         const char *specname;
         char num[33];
         MacroCommand *cmd = &(macro->cmds[j]);

         if(curbatch != cmd->batchid)
         {
            outfile << "  batch" << endl;

            curbatch = cmd->batchid;
         }

         specname = D64_LineSpecNameForNum(cmd->linespecial);
         if(!specname)
         {
            itoa(cmd->linespecial, num, 10);
            specname = num;
         }

         outfile << "    "
                 << specname
                 << " : " 
                 << cmd->linetag 
                 << endl;
      }

      outfile << endl;

      curbatch = -1;
   }

   // done with output file
   outfile.close();
}

//==============================================================================
// 
// Macro Writing
//

//
// WriteSint16
//
// Write a 16-bit little-endian integer to the buffer
//
static void WriteSint16(byte **buffer, int data)
{
   (*buffer)[0] = data & 0xff;
   (*buffer)[1] = (data >> 8) & 0xff;

   *buffer += 2;
}

//
// WriteMacros
//
// Writes the currently loaded set of macros to a binary macros lump.
//
void WriteMacros(const char *fn)
{
   int i;
   size_t size;
   byte *buffer = NULL, *bufptr = NULL;
   ofstream outfile;

   // calculate amount of space needed:
   // * 2 shorts for header
   // * macrocount shorts for macro headers
   // * 3 * (commandcount + 1) shorts for each macro
   size = 2 * (2 + header.macrocount);

   for(i = 0; i < header.macrocount; ++i)
      size += 2 * (3 * (macros[i].commandcount + 1));

   // allocate output buffer
   bufptr = buffer = new byte [size];

   memset((void *)buffer, 0, size);

   // write header

   // note that command counts do NOT include the junk commands between
   // each macro.

   WriteSint16(&bufptr, header.macrocount);
   WriteSint16(&bufptr, header.commandcount);
   
   // write macros
   for(i = 0; i < header.macrocount; ++i)
   {
      int j;
      Macro *macro = &macros[i];

      WriteSint16(&bufptr, macro->commandcount);

      // write commands for this macro - batchid,tag,special
      for(j = 0; j < macro->commandcount; ++j)
      {
         MacroCommand *cmd = &(macro->cmds[j]);

         WriteSint16(&bufptr, cmd->batchid);
         WriteSint16(&bufptr, cmd->linetag);
         WriteSint16(&bufptr, cmd->linespecial);
      }

      // skip junk command (will remain all zeroes)
      bufptr += 3*2;
   }

   // open the output file
   outfile.open(fn, ios::binary);
   
   if(!outfile.is_open())
      throw exception("Could not open macro output file");

   // write buffer to file
   outfile.write((char *)buffer, size);

   // done with output file
   outfile.close();

   // done with buffer
   delete [] buffer;
}

//
// FreeMacros
//
// Deletes all the macros.
//

// EOF

