/*

  Doom 64 Macro Compiler

  Copyright 2008 James Haley

  haleyjd@hotmail.com

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <exception>
#include <fstream>
#include <string>

#include "macrodata.h"
#include "linetable.h"

using namespace std;

//==============================================================================
//
// Script Reading
//

/*

  Script format:

  macro
     batch
        special : tag
        ...
     endbatch
     ...
  endmacro
  ...
*/

// state enumeration for tokenizer
enum
{
   TOKEN_SCAN,
   TOKEN_INWORD,
};

#define NEXTCHAR() *buffer += 1; continue

//
// GetToken
//
// Tokenizer for macro script.
//
static bool GetToken(byte **buffer, string &token)
{
   char c;
   int state = TOKEN_SCAN;

   while((c = **buffer))
   {
      switch(state)
      {
      case TOKEN_SCAN:
         switch(c)
         {
         case ' ':
         case '\t':
         case '\r':
         case '\n':
            // skip whitespace
            NEXTCHAR();
         case ':':
            // a colon is a token all unto its own
            token.append(1, c);
            *buffer += 1;
            return true;
         default:
            // anything else starts a word
            token.append(1, c);
            state = TOKEN_INWORD;
            NEXTCHAR();
         }
         break;

      case TOKEN_INWORD:
         switch(c)
         {
         case ' ':
         case '\t':
         case '\r':
         case '\n':
         case ':':
            // whitespace or colon ends current token
            return true;
         default:
            // anything else is added to the token
            token.append(1, c);
            NEXTCHAR();
         }
         break;

      default:
         throw exception("Internal error: undefined lexer state");
      }
   }

   // eof has been reached
   // if in state word, the last parsed token still needs to be processed.
   // the next call will return false for EOF
   return (state == TOKEN_INWORD);
}

// parser states
enum
{
   STATE_START,    // looking for "macro"
   STATE_INMACRO,  // looking for "batch"
   STATE_INBATCH,  // parsing commands
   STATE_COLON,    // looking for colon
   STATE_CMDARG,   // command argument
   STATE_CMDEND,   // end of command; looking for next command, "batch", or "macro"
};

//
// ReallocMacros
//
static Macro *ReallocMacros(Macro *macros, int nummacros, int nummacrosalloc)
{
   // why does C++ not have a realloc operation? who knows.
   Macro *newMacros = new Macro [nummacrosalloc];

   memset((void *)newMacros, 0, nummacrosalloc * sizeof(Macro));

   if(macros)
   {
      memcpy((void *)newMacros, macros, nummacros * sizeof(Macro));
      delete [] macros;
   }

   return newMacros;
}

//
// AddCommand
//
// Adds a command to the macro.
//
static MacroCommand *AddCommand(Macro *macro)
{
   ++macro->commandcount;

   if(macro->commandcount >= macro->commandsalloc)
   {
      MacroCommand *newCmds;

      // yet more reallocation 9_9
      macro->commandsalloc = macro->commandsalloc ? macro->commandsalloc * 2 : 16;

      newCmds = new MacroCommand [macro->commandsalloc];

      memset((void *)newCmds, 0, macro->commandsalloc * sizeof(MacroCommand));

      if(macro->cmds)
      {
         memcpy((void *)newCmds, macro->cmds, 
                macro->commandcount * sizeof(MacroCommand));
         delete [] macro->cmds;
      }

      macro->cmds = newCmds;
   }

   // return pointer to the new command
   return &(macro->cmds[macro->commandcount - 1]);
}

//
// ParseScript
//
// Parses the macro script from a buffer
//
static void ParseScript(byte *buffer)
{
   string token;
   int state = STATE_START;
   int nummacros = 0;
   int nummacrosalloc = 0;
   int i = 0;
   Macro *lmacros  = NULL;
   Macro *curmacro = NULL;
   MacroCommand *curcommand = NULL;
   int batchNum = 0;
   bool skiptokenread = false;

   while(skiptokenread || GetToken(&buffer, token))
   {
      skiptokenread = false;

      switch(state)
      {
      case STATE_START: // first token must be "macro"
         if(!token.compare("macro"))
         {
            ++nummacros;
            if(nummacros >= nummacrosalloc)
            {
               nummacrosalloc = nummacrosalloc ? nummacrosalloc * 2 : 16;
               lmacros = ReallocMacros(lmacros, nummacros, nummacrosalloc);
            }
            curmacro = &lmacros[nummacros - 1];
            state = STATE_INMACRO;
            // restart batchNum at beginning of a new macro
            batchNum = 0;
         }
         else
            throw exception("Parser error: expected 'macro'");
         break;

      case STATE_INMACRO: // expecting "batch" - a macro cannot be empty
         if(!token.compare("batch"))
         {
            // start a new batch
            batchNum += 10;
            state = STATE_INBATCH;
         }
         else
            throw exception("Parser error: expected 'batch'");
         break;

      case STATE_INBATCH: // looking for start of first command
         if(!token.compare("batch") || !token.compare("macro"))
            throw exception("Parser error: empty command batch not allowed");

         // start a new command
         curcommand = AddCommand(curmacro);

         {
            const char *str = token.c_str();
            char *end = NULL;
            long num;

            num = strtol(str, &end, 0);

            // if it is not a number, try a name
            if(end && *end != '\0')
            {
               num = D64_LineSpecNumForName(str);
               
               if(num < 0)
                  throw exception("Parser error: unknown line special");
            }

            curcommand->batchid = batchNum;
            curcommand->linespecial = num;
         }
         state = STATE_COLON;
         break;

      case STATE_COLON: // expecting colon
         if(token.compare(":"))
            throw exception("Parser error: expected :");

         state = STATE_CMDARG;
         break;

      case STATE_CMDARG: // expecting numeric argument
         if(!token.compare("batch") || !token.compare("macro"))
            throw exception("Parser error: missing command argument");

         {
            const char *str = token.c_str();
            char *end = NULL;
            long num;

            num = strtol(str, &end, 0);

#if 0
            // 06/16/08: Kaiser says the line specials don't take texture
            // numbers after all, so this is disabled for now.

            // if it is not a number, try a texture name
            if(end && *end != '\0')
            {
               num = TextureNumForName(str);

               if(num < 0)
                  throw exception("Parser error: unknown texture name");
            }
#endif
            curcommand->linetag = num;
         }

         state = STATE_CMDEND;
         break;

      case STATE_CMDEND: // expecting next command, batch, or macro
         if(!token.compare("macro")) // start of next macro
            state = STATE_START;
         else if(!token.compare("batch")) // start of next batch
            state = STATE_INMACRO;
         else
            state = STATE_INBATCH; // start of next command

         skiptokenread = true; // we have read the next token already.
         break;

      default:
         throw exception("Internal error: undefined parser state");
      }

      // clear the token buffer string
      if(!skiptokenread)
         token.erase();
   }

   // Reached EOF
   if(state != STATE_CMDEND)
      throw exception("Parser error: unexpected end of file");

   // Really should not happen.
   if(nummacros == 0)
      throw exception("No macros were defined, nothing to do.");

   // set global macro data
   header.macrocount   = nummacros;
   header.commandcount = 0;
   
   for(i = 0; i < nummacros; ++i)
      header.commandcount += lmacros[i].commandcount;

   macros = lmacros;
}

//
// LoadScript
//
// Loads a macro script from a file.
//
void LoadScript(const char *fn)
{
   int length;
   ifstream infile;
   byte *buffer;

   infile.open(fn, ios::binary);

   if(!infile.is_open())
      throw exception("Could not open script input file");

   // get file length
   infile.seekg(0, ios::end);
   length = infile.tellg();
   infile.seekg(0, ios::beg);
   
   if(length <= 0)
      throw exception("Input script file is empty");

   // allocate buffer for file at length + 1
   buffer = new byte [length + 1];

   // read file into buffer
   infile.read((char *)buffer, length);

   // null-terminate the buffer
   buffer[length] = '\0';

   // done with input file
   infile.close();

   // parse the script
   ParseScript(buffer);

   // done with buffer
   delete [] buffer;
}

// EOF

