/*

  Doom 64 Macro Compiler

  Copyright 2008 James Haley

  haleyjd@hotmail.com

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <string>
#include <fstream>
#include <exception>
#include <cstring>

#include "macrodata.h"
#include "textures.h"
#include "utils.h"

using namespace std;

const int NUMTEXTURECHAINS = 701;
static Texture textures;                          // textures in order (dummy)
static Texture *lasttexture = &textures;          // last texture added
static Texture *texturesbyname[NUMTEXTURECHAINS]; // textures by name

// Texture lookup file is a linear list of texture (and flat) names.

//
// TextureNumForName
//
// Returns the number of the texture if found, else -1.
//
int TextureNumForName(const char *name)
{
   int ret = -1;
   unsigned int key = StringHash(name) % NUMTEXTURECHAINS;
   Texture *chain = texturesbyname[key];

   while(chain && strcasecmp(chain->name, name))
      chain = chain->namenext;

   if(chain)
      ret = chain->num;

   return ret;
}

//
// TextureNameForNum
//
// Returns the name of a texture given a number if found, else NULL.
//
const char *TextureNameForNum(int num)
{
   Texture *tx = textures.numnext;
   const char *ret = NULL;

   while(tx && tx->num <= num) // textures are in natural order
   {
      if(tx->num == num)
      {
         ret = tx->name;
         break;
      }

      tx = tx->numnext;
   }

   return ret;
}

//
// AddTexture
//
// Adds a texture to the name hash table and numeric list.
//
static void AddTexture(Texture *tx)
{
   unsigned int key = StringHash(tx->name) % NUMTEXTURECHAINS;

   // add to start of chain
   tx->namenext = texturesbyname[key];
   texturesbyname[key] = tx;

   // add to end of numeric list
   lasttexture->numnext = tx;
   lasttexture = tx;
}

//
// LoadTextures
//
// Loads the texture objects from the texture lookup file.
//
void LoadTextures(const char *fn)
{
   ifstream infile;
   string   buffer;
   int textureNum = 0;

   infile.open(fn);

   if(!infile.is_open())
      throw exception("Failed to open texture lookup file.");

   while(getline(infile, buffer))
   {
      Texture *newTexture = new Texture;
      int nameSize = buffer.length() + 1;

      newTexture->name = new char [nameSize];
      memset((void *)newTexture->name, 0, nameSize);
      strncpy(newTexture->name, buffer.c_str(), nameSize);

      newTexture->num = textureNum++;

      newTexture->namenext = NULL;
      newTexture->numnext  = NULL;

      // link it in
      AddTexture(newTexture);
   }

   // done with input
   infile.close();
}

//
// FreeTextures
//
// Disposes of all Texture objects.
//
void FreeTextures(void)
{
   Texture *curtx = textures.numnext;

   // use numeric chain to delete all objects
   while(curtx)
   {
      Texture *next = curtx->numnext;

      delete [] curtx->name;
      delete curtx;

      curtx = next;
   }

   // reset all globals
   textures.numnext = NULL;
   lasttexture = &textures;
   memset((void *)texturesbyname, 0, NUMTEXTURECHAINS * sizeof(Texture *));
}

// EOF

