/*

  Doom 64 Macro Compiler

  Copyright 2008 James Haley

  haleyjd@hotmail.com

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <iostream>
#include <exception>
#include <cstdlib>

#include "macrodata.h"

using namespace std;

// action that program will perform
enum
{
   ACTION_COMPILESCRIPT,
   ACTION_DECOMPILELUMP,
   ACTION_NORMALIZELUMP
};

static int action;
static const char *infile;
static const char *outfile;

//
// ShowUsage
//
// Prints usage info and exits the program.
//
static void ShowUsage(void)
{
   cerr << "Usage:\n"
        << " macrocmp <infile> -c [<outfile>] : compile script\n"
        << " macrocmp <infile> -d [<outfile>] : decompile lump\n"
        << " macrocmp <infile> -n [<outfile>] : normalize lump"
        << endl;

   exit(0);
}

//
// ScanCmdLine
//
// Looks for options on the command line.
// The following options are possible:
//
// <infilename> -c [<outfilename>] : script will be compiled to binary lump
// <infilename> -d [<outfilename>] : binary lump will be decompiled
// <infilename> -n [<outfilename>] : binary lump will be normalized
//
static void ScanCmdLine(int argc, const char *const argv[])
{
   const char *option;

   if(argc < 3)
      ShowUsage();

   infile = argv[1];
   option = argv[2];

   // figure out the option
   if(strlen(option) < 2)
      ShowUsage();

   switch(option[1])
   {
   case 'C':
   case 'c':
      action = ACTION_COMPILESCRIPT;
      break;
   case 'D':
   case 'd':
      action = ACTION_DECOMPILELUMP;
      break;
   case 'N':
   case 'n':
      action = ACTION_NORMALIZELUMP;
      break;
   default:
      ShowUsage();
   }

   if(argc >= 4)
      outfile = argv[3];
   else
   {
      if(action == ACTION_COMPILESCRIPT || action == ACTION_NORMALIZELUMP)
         outfile = "out.lmp";
      else
         outfile = "out.txt";
   }
}

//
// main program
//
int main(int argc, const char *const argv[])
{
   try
   {
      // get command line options
      ScanCmdLine(argc, argv);

      cout << "macrocmp - by James 'Quasar' Haley <haleyjd@hotmail.com>\n\n"
           << "This program is available under the terms of the GNU Genera Public\n"
           << "License. See the file 'COPYING' for full details.\n" << endl;

      // perform the instructed action
      switch(action)
      {
      case ACTION_COMPILESCRIPT:
         cout << "Compiling script " << infile << " to " << outfile << endl;
         LoadScript(infile);
         WriteMacros(outfile);
         break;
         
      case ACTION_DECOMPILELUMP:
         cout << "Decompiling lump " << infile << " to " << outfile << endl;
         LoadMacros(infile);
         DecompileMacros(outfile);
         break;

      case ACTION_NORMALIZELUMP:
         cout << "Normalizing lump " << infile << " to " << outfile << endl;
         LoadMacros(infile);
         WriteMacros(outfile);
         break;
      }
   }
   catch(exception &e)
   {
      cerr << "Error: " << e.what() << endl;
   }

   return 0;
}

// EOF


